/// AudioSample.h
///
/// (c) 2001-2006 by Bastian Spiegel <bs@tkscript.de>
///     - Distributed under terms of the Lesser GNU General Public License (LGPL).
///       See COPYING and <http://www.gnu.org/licenses/licenses.html#LGPL> for further information.
///

#ifndef __TKS_AUDIOSAMPLE_H__
#define __TKS_AUDIOSAMPLE_H__

//class TKSAVIFileAudio;

YG("audio");

#define cAFloat28(a) ((sS8)((a)*127.99))
#define cAFloat216(a) ((sS16)((a)*32767.99))
#define cAFloat224(a) ((sS32)((a)*16777215.99))
#define cAFloat232(a) ((sS32)((a)*((((sF32)(1<<31))-0.01)*2.0)))


YC class _AudioSample : public YAC_Object {
public:
  /// Sample data types
  enum _stypes {
    /// Signed 8 bit samples
    Sample8=1,
    /// Signed 16 bit samples
    Sample16=2,
    /// Signed 24 bit samples
    Sample24=3,
    /// Signed 32 bit samples
    Sample32=4,
    /// 32bit Float samples
    SampleFloat=5,
    /// 64bit Double samples
    SampleDouble=6
  };

  /// Buffer types
  enum _atypes {

    /// Time domain virtual buffers represent parts of another time domain buffer. 
    /// The buffer layout is: Vertical: Subwave , Horizontal: Time (sample data)
    /// If the vertical (y) size of the buffer is >1 the buffer is considered
    /// to consist of multiple subwaves that are accessable by the y index.
    /// This can be used for wavetable/granular synthesis.
    ///
    /// The memory area is of unknown type
    ///
    TD_VIRTUAL=0,

    /// Frequency domain virtual buffer represent parts of another frequency domain
    /// buffer
    /// The buffer layout is: Vertical: time , Horizontal: Freqency spectrum
    /// You cannot convert a sample with subwaves into frequency domain.
    /// The memory area is of unknown type
    ///
    FD_VIRTUAL=1,

    /// Time domain Audio buffers are located in audio driver space. 
    /// The buffer layout is: Vertical: 1, Horizontal: Time (samples data)
    /// What you write is what you hear.
    //
    TD_AUDIO=2,

    /// Frequency domain Audio buffers are locate in audio driver space.
    /// Actually I do not know about any soundcard that is able to do
    /// an inverse DFT in hardware. 
    /// Can be used by audio drivers with integrated DSPs ?
    ///
    FD_AUDIO=3,

    /// Time domain raw buffers are located in main memory.
    /// Fast to access by CPU but they need to be copied to an AUDIO buffer 
    /// to aurealize them 8)
    ///
    TD_RAW=4,

    /// Frequency domain raw buffers are located in main memory.
    /// The buffer layout is : Vertical: time , Horizontal: freq spectrum
    ///
    FD_RAW=5
  };


public:
  /// Pointer to audio buffer
  Bitmap *buffer;

  /// True if the buffer is allocated/freed by this class
  sBool own_buffer;
  
  /// Sample-Type of audio buffer
  /// o
  int stype;

  int type;

  /// Number of interleaved channels
  /// 1=Mono, 
  /// 2=Stereo, 
  /// 3=?, 
  /// 4=Quad, 
  /// 5=Surround (seperated frontleft, frontright, rearleft, rearright, bass)
  /// >5 = multichannel :)
  /// Default=0
  int channels;

  /// Sample frequency in Hz (number of samples/sec)
  /// Default=0
  sF32 frequency;

  /// Sample base frequency in Hz
  sF32 base_frequency;

  /// Sample offset (number of samples)
  /// Default=0
  int offset;

  /// Sample length (number of samples)
  /// This must NOT match the buffer (horizontal/vertical) size
  /// Default=0
  int length;

  /// Sample repeat offset (number of samples)
  /// Default=0
  int repeat_offset;

  /// Sample repeat length (number of samples)
  /// Default=0
  int repeat_length;

   /// FFT Window size for FD_* buffers.
  int fft_m;
 
  /// reference to avi audio stream class if in use
  //TKSAVIFileAudio *avifileaudio;

public:
  _AudioSample();
  ~_AudioSample();

  YAC(_AudioSample);

  /// Allocate new audio buffer
  /// @param s Sample type (see enum _stypes). One of Sample[8,16,32,64]
  /// @param t Type (see enum _atypes). Must be TD_RAW or FD_RAW.
  /// If you allocate a frequency domain (FD_) buffer, every line in the
  /// buffer represents the frequency spectrum for the time domain sample nr y.
  /// @param l Number of samples (/subwave!)
  /// @param f Frequency. Does not affect the buffer size.
  /// @param i Number of interleaved channels
  /// @param sw Number of waves (if type != FD_*).
  /// @param fftwin FFT window size (if type==FD_*) (=>Number of frequencies/ sample)
  /// Typical values are 1,2,3,4,5,6,7,8,9,10. A window size of fftwin will
  /// result in n=((1<<fftwin)/2) frequencies 
  /// ( f= (i*s)/(1<<(fftwin+1)) ; i=1..n-1, s= 1..44100 )
  /// ( i= (f*(1<<(fftwin+1)))/s ; i=1..n-1, s= 1..44100 )
  ///
  /// @return True if buffer was allocated. False if not
  sBool allocBuffer(int s, int t, int l, sF32 f=44100.0, int i=1, int sw=1, int fftwin=0  );

  /// Delete buffer if it is owned by this object (own_buffer flag)
  void deleteBuffer();
  
  /// Use auxiliary buffer (shared with other object)
  /// @param b Pointer to audiobuffer 
  void setBuffer(Bitmap *b);

  /// Generate a sinus signal with frequency <b>freq</b> using a
  /// samplerate of <b>frate</b> (mono)
  /// @param s Sampletype
  /// @param l Number of samples
  /// @freq Sinus frequency
  /// @frate Signal sample rate
  void generateSinus(int s, int l, float freq=440, float frate=44100);

  /// Generate a 0 signal
  /// @param s Sampletype
  /// @param l Number of samples
  /// @param c Number of channels
  /// @param frate Signal sample rate
  void generateSilence(int s, int l, int c, float frate=44100.0);

  /// Generate a pulse signal with frequency <b>freq</b> using a
  /// samplerate of <b>frate</b> (mono).
  /// The generated pulse will have a pulsewidth of <b>pw</b> degrees.
  ///
  /// @param s Sampletype
  /// @param l Number of samples
  /// @param pw Pulsewidth in 0..2PI
  /// @param freq Pulse frequency
  /// @param frate Signal sample rate
  ///
  void generatePulse(int s, int l, float pw, float freq=440, float frate=44100);


  /// Normalize samples. 
  /// This will multiply all samples with MAX_RANGE/max_used_amplitude
  void normalize();

  /// Get index of maximum amplitude in time domain sample.
  /// @param sw Subwave # (0..buffer->s.y-1)
  /// @return Index ( x-position in buffer )
  int getTDMax(int sw=0);

  sBool loadAIFFFile(sBool _local, YAC_String *fname);

  void getNextLoop(sU16 *_loopindex, sU32 *_offset, sU32 *_length);

  void YAC_CALL yacOperator(sSI, YAC_Object*, YAC_Value *_r);

  YM sSI  _allocBuffer      (sSI _l, sF32 _f, sSI _ch, sSI _sw);
  YM void _freeBuffer       (void);
  YM void _visitBuffer      (YAC_Object *);
  YM void _normalize        (void);
  YM sSI  _loadAIFF         (YAC_String *_name);
  YM sSI  _loadLocalAIFF    (YAC_String *_name);
  YM sSI  _getChannels      (void);
  YM sF32 _getFrequency     (void);
  YM sSI  _getOffset        (void);
  YM void _setOffset        (sSI);
  YM sSI  _getLength        (void);
  YM void _setLength        (sSI);
  YM sSI  _getRepeatOffset  (void);
  YM void _setRepeatOffset  (sSI);
  YM sSI  _getRepeatLength  (void);
  YM void _setRepeatLength  (sSI);
  YM sSI  _copyBufferFrom   (YAC_Object *_sample);
  YM sF32 _getBaseFrequency (void);
  YM void _setBaseFrequency (sF32);
  YM void _setFrequency     (sF32);
  YM sSI  _convertChannels  (sSI ch);
};


#endif
