/// HAL.cpp
///
/// (c) 2001-2006 by Bastian Spiegel <bs@tkscript.de>
///     - Distributed under terms of the Lesser GNU General Public License (LGPL).
///       See COPYING and <http://www.gnu.org/licenses/licenses.html#LGPL> for further information.
///

#include "static_tksdl.h"

#include <stdlib.h>

#define YAC_PRINTF defined
#include <yac.h>

#ifdef YAC_WIN32
#include <windows.h>
#include <ddraw.h>
#endif
#include "HAL.h"
#include "Joystick.h"
#include "Key.h"
#include "Mouse.h"

extern YAC_Host *yac_host;
extern sUI clid_X_IVector;
extern sUI clid_X_Vector;
extern sUI clid_Joystick;
extern sUI clid_Key;

#include <tks-opengl.h>

#include <SDL/SDL_keysym.h>
#include <SDL/SDL_net.h>
#include <SDL/SDL_joystick.h>
#include <SDL/SDL_syswm.h>

static SDL_SysWMinfo syswminfo;

#include "Bitmap.h"
#include "HAL.h"

extern _HAL *tks_HAL;

extern void*sc_context; // script execution context for HAL callbacks (onReopen, onKeyboard, onMouse, ...)

// ---- SDL "scrap" clipboard support (x11/qnx/win32)
#ifdef TKSDL_SCRAP_CLIPBOARD
extern int init_scrap(void);
#endif


sUI GetClassID_IVector(void) {
   if(clid_X_IVector==(sUI)-1)
      clid_X_IVector=yac_host->yacGetClassIDByName("IVector");
   return clid_X_IVector;
}

sUI GetClassID_Vector(void) {
   if(clid_X_Vector==(sUI)-1)
      clid_X_Vector=yac_host->yacGetClassIDByName("Vector");
   return clid_X_Vector;
}



//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//~                                                                ~
//~ HAL                                                            ~
//~                                                                ~
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
_HAL::_HAL(void) {
   b_unicode=0;
   b_running=0;
   b_graphics_active=0;
   b_reopen=0;
   b_need_redraw=0;
   b_event_polling=1;
   b_double_buffer = 1;
   b_redrawing = 0;
   max_frame_deltatime=0; //(1000.0f/99999999.0); 20;
   
   int i;
   for(i=0; i<TKS_HAL_NUMSIGNALS; i++)
   {
      signal_funs[i]=0;
   }
   mouse.show=1;
   mouse.def_auto_hide_timeout=1500;
   mouse.auto_hide_timeout=0;
   mouse.pointer_mode=TKS_MOUSE_POINTER_MOVE;
   
   joysticks.enabled=0;
   joysticks.num=0;
   joysticks.elements=0;
   
   redraw_flag=TKS_REDRAW_ALL;
   
   win_sx=0;
   win_sy=0;
   fs_sx=1024;
   fs_sy=768;
   fs_z=32;
   view_sx=0;
   view_sy=0;
   view_z=0;
   view_flags=0;
   tks_flags=SDL_RESIZABLE;
   
#ifdef YAC_WIN32
   lpDD=0;
   dw_monitorfreq=0;
#endif
   
   _initPickingColors();
   
   audio.dsp_frequency=44100.0f; // overwritten by AudioDevice::openDSP()
   
   fps.tick_buffer=0;
   fps.tick_interval=1000.0f/60;
   fps.limit_fps=0;
   
   mouse.last_absx=999;
   mouse.last_absy=999;
}

_HAL::~_HAL() {
#ifdef YAC_WIN32
   if(lpDD)
   {
      lpDD->Release();
      lpDD=0;
   }
#endif
} 

void YAC_CALL _HAL::yacGetSignalStringList(YAC_String *_sig) {
   // ---- the number represents the encoded argument type identifiers used by this function
   // ---- 1=int, 2=float, 3=object. bits 0+1 represent argument #0, bits 2+3 represent argument #1 and so on
   _sig->visit("onSignal:1 onKeyboard:3 onMouse:85 onJoyAxisMotion:21 onJoyButton:21 onJoyHatMotion:21 onJoyBallMotion:85 onDraw:0 onTick:0 onExit:0 onEvent:3 onReopen:0 onResize:0 onExpose:0 onApplicationActive:1 onMouseFocusActive:1 onIdle:0");
}

void YAC_CALL _HAL::yacRegisterSignal(sUI _id, void *_fun) {
   if(_id<TKS_HAL_NUMSIGNALS)
   {
      signal_funs[_id]=_fun;
   }
}

sSI _HAL::_getNumJoysticks(void) {
   return joysticks.num;
}

void _HAL::_getJoystick(sSI _idx, YAC_Value *_r) {
   if(((sU32)_idx)<((sU32)joysticks.num))
   {
      _Joystick *j=&((_Joystick*)joysticks.elements)[_idx];
      YAC_RETO(j, 0);
   }
   else
   {
      YAC_RETO(0, 0);
   }
}

void _HAL::_updateJoysticks(void) {
   if(joysticks.enabled)
      ::SDL_JoystickUpdate();
}


void _HAL::_triggerEvent(YAC_Object *_s) {
   YAC_String *s;
   if(YAC_CHK(_s, YAC_CLID_STRING))
   {
      s=(YAC_String*)_s;
   }
   else
   {
      if(YAC_CHK(_s, YAC_CLID_EVENT))
      {
         YAC_Event *a=(YAC_Event *)_s;
         s=&a->data;
      }
      else
         s=0;
   }
   if(s)
   {
      void *f=signal_funs[TKS_SIG_HALUSEREVENT];
      if(f)
      {
         YAC_Value args[1];
         args[0].initString(s, 0);
         yac_host->yacEvalFunction(sc_context, f, 1, args);
      }
   }
}

void _HAL::_sendUserEvent(YAC_Object *_s) {
   if(YAC_CHK(_s, YAC_CLID_STRING))
   {
      SDL_Event user_event;
      YAC_String *ns=YAC_New_String();
      ns->yacCopy((YAC_String*)_s);
      user_event.type=SDL_USEREVENT;
      user_event.user.code=TKS_EVENT_ONACTION;
      user_event.user.data1=(void *)ns;
      user_event.user.data2=(void *)::SDL_GetTicks();
      ::SDL_PushEvent(&user_event);
   }
   else
   {
      if(YAC_CHK(_s, YAC_CLID_EVENT))
      {
         YAC_Event *a=(YAC_Event *)_s;
         SDL_Event user_event;
         YAC_String *ns=YAC_New_String();
         ns->yacCopy(&a->data);
         user_event.type=SDL_USEREVENT;
         user_event.user.code=TKS_EVENT_ONACTION;
         user_event.user.data1=(void *)ns;
         user_event.user.data2=(void *)a->time_stamp;
         ::SDL_PushEvent(&user_event);
      }
   }
   // xxx else toString..
}

void _HAL::_initPickingColors(void) {
   sU16 i;
   sU8 *d=picking.gb;
   for(i=0; i<1024; i++)
   {
      *d++=0;
      *d++=0;
   }
}

sU8 *_HAL::allocPickingColor(void) {
   // ---- allocate unique picking color which is also usuable on 555 16bit displays
   sU16 i;
   yacmemptr d;
   d.u1=picking.gb;
   for(i=1; i<1024; i++)
   {
      if(!*d.u2)
      {
         d.u1[0]=((i>>5)&0x1f)<<3; // g
         d.u1[1]=(i&0x1f)<<3;      // b
         return d.u1;
      }
      d.u2++;
   }
   return 0;
}

sUI _HAL::_allocPickingColor(void) {
   sU8 *r=allocPickingColor();
   if(r)
      return *((sUI*)r);
   else
      return 0;
}

void _HAL::_enableKeyRepeat(sSI _d, sSI _r) {
   ////yac_host->printf("enablekeyRepeat(_d=%i _r=%i=%i)\n", _d, _r, SDL_EnableKeyRepeat(_d, _r));
   ::SDL_EnableKeyRepeat(_d, _r);
}

void _HAL::openJoysticks(void) {
   // ---- do not open twice ----
   if(!joysticks.enabled)
   {
      joysticks.enabled=1;
      joysticks.num=::SDL_NumJoysticks();
      int i;
      _Joystick *jo=new _Joystick[joysticks.num]();
      joysticks.elements=(void*)jo;
      for(i=0; i<joysticks.num; i++)
      {
         jo[i].class_ID=clid_Joystick;
         jo[i].open(i);
      }
   }
}

void _HAL::closeJoysticks(void) {
   if(joysticks.enabled)
   {
      joysticks.enabled=0;
      _Joystick*jo=(_Joystick*) joysticks.elements;
      if(jo)
      {
         int i;
         for(i=0; i<joysticks.num; i++)
         {
            jo[i].close();
         }
         delete [] jo;
         joysticks.elements=0;
         joysticks.num=0;
      }
   }
}


void _HAL::_exitEventLoop(void) {
   b_running=0;
   if(b_graphics_active)
   {
      showCursor(1);
   }
}

void _HAL::tick(void) {
   void *f=signal_funs[TKS_SIG_HALTICK];
   if(f) yac_host->yacEvalFunction(sc_context, f, 0, 0);
}


void _HAL::draw(void) {
   if(!now) return; // wait for first tick
   if(!fps.last_frametime)
   {
      fps.last_frametime=now;
      fps.precision=1.0f;
   }
   else
   {
      sSI dt=now-fps.last_frametime;
      if(dt<0) dt=0; // wrap-around
      if(fps.limit_fps)
      {
         sF32 limit=1000.0f/fps.limit_fps; // desired tickinterval
         if(dt<limit)
            ::SDL_Delay((sUI)(limit-dt));
         now=::SDL_GetTicks();
         dt=now-fps.last_frametime;
      }
      sF32 np=((sF32)dt)/fps.tick_interval;
      fps.precision=fps.precision+(np-fps.precision)*0.75f;
      fps.current_fps=1000.0f/((sF32)dt);
      fps.last_frametime=now;
      fps.total_time+=dt;
      fps.average_fps=((sF32)fps.total_time)/((sF32)fps.num_frames_rendered);
      fps.average_fps=1000.0f/fps.average_fps;
   }
   
   if(!fps.last_secondtime)
      fps.last_secondtime=now;
   else
   {
      sSI dt=now-fps.last_secondtime;
      if(dt<0)
         dt=0; // wrap-around
      if(dt>=1000&&fps.second_frames_rendered)
      {
         fps.real_fps=((sF32)dt)/((sF32)fps.second_frames_rendered);
         fps.real_fps=(1000.0f/fps.real_fps);
         fps.last_secondtime=now;
         fps.second_frames_rendered=0;
      }
   }
   fps.second_frames_rendered++;
   fps.num_frames_rendered++;
   
   if((redraw_flag==TKS_REDRAW_ALL))
   {
      void *f=signal_funs[TKS_SIG_HALDRAW];
      if(f)
      {
         beginPaint();
         yac_host->yacEvalFunction(sc_context, f, 0, 0);
         endPaint();
      }
   }
   else
   {
      if(b_need_redraw)
      {
         b_need_redraw=0;
         void *f=signal_funs[TKS_SIG_HALDRAW];
         if(f)
         {
            beginPaintDirty();
            yac_host->yacEvalFunction(sc_context, f, 0, 0);
            endPaint();
         }
      }

      // ---- Call onIdle script function (e.g. to process timers etc)
      void *f=signal_funs[TKS_SIG_HALIDLE];
      if(f)
      {
         yac_host->yacEvalFunction(sc_context, f, 0, 0);
      }
      ::SDL_Delay(10);
   }
}


sBool _HAL::loopDelay(void) { 
   now=::SDL_GetTicks();
   
   if(!last_time)
   {
      last_time=now;
      return 0;
   }
   else
   {
      if((last_time+((sU32)(fps.tick_buffer*fps.tick_interval)))<=now)
      {
         last_time=now;
         sU16 i;
         for(i=0; i<fps.tick_buffer; i++) tick();
         return 1;
      }
      else
      {
         if((last_time+((sU32)fps.tick_interval))<=now)
         {
            while(last_time<now)
            {
               tick();
               last_time+=((sU32)fps.tick_interval);
               if((last_time+fps.tick_interval*0.75f)>=now)
                  return 1;
            }
            return 1;
         }
         else return 0;
      }
   }
}

void _HAL::initFPS(void) {
   fps.num_frames_rendered=0;
   fps.total_time=0;
   fps.last_frametime=0;
   fps.average_fps=0.0f; 
   fps.current_fps=0.0f;
   fps.real_fps=0.0f;
   fps.second_time=0;
   fps.last_secondtime=0;
   fps.second_frames_rendered=0;
   fps.precision=1.0;
}

void _HAL::callKeyboardHandler(sU32 _vkey, sU32 _vmod, sUI _unicode) {
   sU32 bp_vkey=_vkey;
   if(bp_vkey)
   {
      _Key k;
      k.class_ID=clid_Key;
      k.vkey=bp_vkey;
      k.vmod=_vmod;
      k.unicode=_unicode;
      void *f=signal_funs[TKS_SIG_HALKEYBOARD];
      if(f) // ---- invoke keyboard handler if present.. ----
      {
         YAC_Value args[1];
         args[0].initObject(&k, 0);
         yac_host->yacEvalFunction(sc_context, f, 1, args);
      }
   }
}

void _HAL::callExit(void) {
   if(signal_funs[TKS_SIG_HALEXIT])
      yac_host->yacEvalFunction(sc_context, signal_funs[TKS_SIG_HALEXIT], 0, 0);
}

void _HAL::callResize(void) {
   if(signal_funs[TKS_SIG_HALRESIZE])
      yac_host->yacEvalFunction(sc_context, signal_funs[TKS_SIG_HALRESIZE], 0, 0);
   callReopen();
}

void _HAL::callExpose(void) {
   if(signal_funs[TKS_SIG_HALEXPOSE])
      yac_host->yacEvalFunction(sc_context, signal_funs[TKS_SIG_HALEXPOSE], 0, 0);
}

void _HAL::callApplicationActive(sSI _gain) {
   if(signal_funs[TKS_SIG_HALAPPLICATIONACTIVE])
   {
      YAC_Value args[1];
      args[0].initInt(_gain);
      yac_host->yacEvalFunction(sc_context, signal_funs[TKS_SIG_HALAPPLICATIONACTIVE], 1, args);
   }
}

void _HAL::callMouseFocusActive(sSI _gain) {
   if(signal_funs[TKS_SIG_HALMOUSEFOCUSACTIVE])
   {
      YAC_Value args[1];
      args[0].initInt(_gain);
      yac_host->yacEvalFunction(sc_context, signal_funs[TKS_SIG_HALMOUSEFOCUSACTIVE], 1, args);
   }
}

void _HAL::callReopen(void) {
   if(signal_funs[TKS_SIG_HALREOPEN]) 
   {
      tks_HAL->b_reopen=1;
      yac_host->yacEvalFunction(sc_context, signal_funs[TKS_SIG_HALREOPEN], 0, 0);
      tks_HAL->b_reopen=0;
   }
}

void _HAL::callOnJoyAxisMotion(sSI _dev, sSI _axis, sSI _val) {
   void *f=signal_funs[TKS_SIG_HALJOYAXISMOTION];
   if(f)
   {
      YAC_Value args[3];
      args[0].initInt(_dev);
      args[1].initInt(_axis);
      args[2].initInt(_val);
      yac_host->yacEvalFunction(sc_context, f, 3, args);
   }
}

void _HAL::callOnJoyButton(sSI _dev, sSI _button, sSI _pressed) {
   void *f;
   if( (f=signal_funs[TKS_SIG_HALJOYBUTTON]) )
   {
      YAC_Value args[3];
      args[0].initInt(_dev);
      args[1].initInt(_button);
      args[2].initInt(_pressed);
      yac_host->yacEvalFunction(sc_context, f, 3, args);
   }
}

void _HAL::callOnJoyHatMotion(sSI _dev, sSI _hat, sSI _state) {
   void *f;
   if( (f=signal_funs[TKS_SIG_HALJOYHATMOTION]) )
   {
      YAC_Value args[3];
      args[0].initInt(_dev);
      args[1].initInt(_hat);
      args[2].initInt(_state);
      yac_host->yacEvalFunction(sc_context, f, 3, args);
   }
}

void _HAL::callOnJoyBallMotion(sSI _dev, sSI _hat, sSI _dx, sSI _dy) {
   void *f;
   if( (f=signal_funs[TKS_SIG_HALJOYBALLMOTION]) )
   {
      YAC_Value args[4];
      args[0].initInt(_dev);
      args[1].initInt(_hat);
      args[2].initInt(_dx);
      args[3].initInt(_dy);
      yac_host->yacEvalFunction(sc_context, f, 4, args);
   }
}

sBool _HAL::processSDLEvent(sBool _poll) {
   ////yac_host->printf("processSDLEvent\n");
   SDL_Event ev;
   int r;
   if(_poll) r=::SDL_PollEvent(&ev);
   else r=::SDL_WaitEvent(&ev);

   if(r)
   {
      //yac_host->printf("[...] got SDL event. type=%i\n", ev.type);
      sU32 vkey=0;
      sU32 vmod=0;
      switch(ev.type)
      {
      case SDL_JOYAXISMOTION:
         callOnJoyAxisMotion(ev.jaxis.which, ev.jaxis.axis, ev.jaxis.value);
         break;
         
      case SDL_JOYBUTTONDOWN:
      case SDL_JOYBUTTONUP:
         callOnJoyButton(ev.jbutton.which, ev.jbutton.button, ev.jbutton.state==SDL_PRESSED);
         break;
         
      case SDL_JOYHATMOTION:
         callOnJoyHatMotion(ev.jhat.which, ev.jhat.hat, ev.jhat.value);
         break;
         
      case SDL_JOYBALLMOTION:
         callOnJoyBallMotion(ev.jball.which, ev.jball.ball, ev.jball.xrel, ev.jball.yrel);
         break;
         
      case SDL_VIDEORESIZE:
         {
            tks_HAL->openView(ev.resize.w, ev.resize.h, 0/**z**/, 0/**fs**/, 1/**resize**/);
            tks_HAL->callResize();
         }
         break;
         
      case SDL_USEREVENT:
         if(ev.user.code==TKS_EVENT_ONACTION)
         {
            if(ev.user.data1)
            {
               // ---- deliver delayed onAction event ----
               _triggerEvent((YAC_Object*)ev.user.data1);
               delete (YAC_Object*)ev.user.data1;
               ev.user.data1=0;
            }
         }
         break;
      case SDL_ACTIVEEVENT:
         if(ev.active.state==SDL_APPACTIVE)
         {
            if(ev.active.gain)
            {
               b_graphics_active=1;
               // ...
            }
            callApplicationActive((sSI)ev.active.gain);
         }
         else
         {
            if(ev.active.state==SDL_APPMOUSEFOCUS)
            {
               if(ev.active.gain)
               {
                  if(mouse.pointer_mode==TKS_MOUSE_POINTER_ENTER)
                  {
                     mouse.auto_hide_timeout=mouse.def_auto_hide_timeout;
                     mouse.show=TKS_MOUSE_SHOW_ENABLE;
                  }
               }
               callMouseFocusActive((sSI)ev.active.gain);
            }
         }
         
         /*else
         {
         yac_host->printf("mouse leave\n");
      }*/
         break;
         
      case SDL_VIDEOEXPOSE:
         updateUnicode();
         b_graphics_active=1;
         callExpose();
         break;

      case SDL_KEYDOWN:
         vkey|=VKEY_PRESSED;
         
      case SDL_KEYUP:
         vkey|=ev.key.keysym.sym;
         vmod=ev.key.keysym.mod&~KMOD_NUM;
         ///
         ///yac_host->printf("vkey=%08x vmod=%08x unicode=%i enableUNICODE=%i.\n", vkey, vmod, (sUI)ev.key.keysym.unicode, ::SDL_EnableUNICODE(-1));
         callKeyboardHandler(vkey, vmod, (sUI)ev.key.keysym.unicode);

#ifdef YAC_LINUX
         // Fixes a bug witnessed under linux where the unicode field is always 0 until another application
         // has been focused and then the sdl window is activated again
         updateUnicode();
#endif
         break;
         
      case SDL_MOUSEBUTTONUP:
         switch(ev.button.button)
         {
         case SDL_BUTTON_LEFT:
            mouse.cstate&=~TKS_MOUSE_LBUTTON;
            mouse.nstate=TKS_MOUSE_LBUTTON;
            break;
         case SDL_BUTTON_RIGHT:
            mouse.cstate&=~TKS_MOUSE_RBUTTON;
            mouse.nstate=TKS_MOUSE_RBUTTON;
            break;
         case SDL_BUTTON_MIDDLE:
            mouse.cstate&=~TKS_MOUSE_MBUTTON;
            mouse.nstate=TKS_MOUSE_MBUTTON;
            break;
         case SDL_BUTTON_WHEELUP:
            mouse.cstate&=~TKS_MOUSE_WHEELUP;
            mouse.nstate=TKS_MOUSE_WHEELUP;
            break;
         case SDL_BUTTON_WHEELDOWN:
            mouse.cstate&=~TKS_MOUSE_WHEELDOWN;
            mouse.nstate=TKS_MOUSE_WHEELDOWN;
            break;
         }
         callMouseHandler(ev.button.x, ev.button.y, mouse.cstate, mouse.nstate);
         break;
         
     case SDL_MOUSEBUTTONDOWN:
         switch(ev.button.button)
         {
         case SDL_BUTTON_LEFT:
            mouse.cstate|=TKS_MOUSE_LBUTTON;
            mouse.nstate=TKS_MOUSE_LBUTTON;
            break;
         case SDL_BUTTON_RIGHT:
            mouse.cstate|=TKS_MOUSE_RBUTTON;
            mouse.nstate=TKS_MOUSE_RBUTTON;
            break;
         case SDL_BUTTON_MIDDLE:
            mouse.cstate|=TKS_MOUSE_MBUTTON;
            mouse.nstate=TKS_MOUSE_MBUTTON;
            break;
         case SDL_BUTTON_WHEELUP:
            mouse.cstate|=TKS_MOUSE_WHEELUP;
            mouse.nstate=TKS_MOUSE_WHEELUP;
            break;
         case SDL_BUTTON_WHEELDOWN:
            mouse.cstate|=TKS_MOUSE_WHEELDOWN;
            mouse.nstate=TKS_MOUSE_WHEELDOWN;
            break;
         }
         callMouseHandler(ev.button.x, ev.button.y, mouse.cstate, mouse.nstate);
         break;
            
     case SDL_MOUSEMOTION:
            /*if(!mouse.grabbed && mouse.warp_sy)
            {
            mouse.warp_sx=0;
            mouse.warp_sy=0;
            mouse.x=mouse.absx;
            mouse.y=mouse.absy;
            }
            else
               {*/

        ///yac_host->printf("ev.motion=(%i, %i)\n", ev.motion.x, ev.motion.y);
               
        callMouseHandler(ev.motion.x, ev.motion.y, mouse.cstate, 0);
        if(!mouse.grabbed && mouse.pointer_mode==TKS_MOUSE_POINTER_MOVE)
        {
           mouse.auto_hide_timeout=mouse.def_auto_hide_timeout;
           mouse.show=TKS_MOUSE_SHOW_ENABLE;
        }
        else
        {
#ifdef YAC_WIN32
           // workaround to keep mouse pointer hidden
           //TKS_SHOWCURSOR(mouse.show);
#endif
        }
        break;
               
     case SDL_QUIT:
        b_running = 0;
        break;

     default:
        break;
     }
     return 1;
   }
   return 0;
}

void _HAL::resetMouse(void) {
   mouse.absx    = 0;
   mouse.absy    = 0;
   mouse.dx      = 0;
   mouse.dy      = 0;
   mouse.cstate  = 0;
   mouse.nstate  = 0;
   mouse.grabbed = 0;
   mouse.warp_sx = 0;
   mouse.warp_sy = 0;
}

void _HAL::_eventLoop(void) {
   static sU32 last_ticks=0;
  
   b_running = 1;
   last_time = 0;
   now       = 0;

   resetMouse();
   
   while(b_running)
   {
      b_running=yac_host->yacRunning();
      if(b_graphics_active)
      {
         // invoke trace function, redraws rt-graphics screen 
         
         if(loopDelay())//&&b_graphics_active)
         {
            if(mouse.show==TKS_MOUSE_SHOW_ENABLE)
            {
               showCursor(1);
            }
            else
            {
               if(mouse.show==TKS_MOUSE_SHOW_DISABLE)
               {
                  showCursor(0);
               }
            }
            
            if( b_running )
            {
               if(processSDLEvent(b_event_polling))
               {
                  while( b_running && processSDLEvent(1) );
               }
               
               sU32 delta_ticks;
               if(!last_ticks)
               {
                  last_ticks=::SDL_GetTicks();
                  delta_ticks=0;
               }
               else
               {
                  delta_ticks=last_ticks;
                  last_ticks=::SDL_GetTicks();
                  delta_ticks=last_ticks-delta_ticks;
               }
               if(mouse.auto_hide_timeout)
               {
                  mouse.auto_hide_timeout-=delta_ticks;
                  if(((sSI)mouse.auto_hide_timeout)<=0)
                  {
                     mouse.auto_hide_timeout=0;
                     mouse.show=TKS_MOUSE_SHOW_DISABLE;
                  }
               }
               if(mouse.warp_sx)
               {
                  ::SDL_WarpMouse(mouse.warp_sx, mouse.warp_sy);
                  mouse.warp_sx=0;
               }
            }
         }
         draw();
      }
      else
      {
         // ---- no graphics output ----
         ::SDL_Delay(10);
      }
   }
   if(signal_funs[TKS_SIG_HALEXIT])
      yac_host->yacEvalFunction(sc_context, signal_funs[TKS_SIG_HALEXIT], 0, 0);
}

void _HAL::callMouseHandler(sSI _mx, sSI _my, sSI _cstate, sSI _nstate) {
   mouse.cstate=(sU16)_cstate;
   mouse.nstate=(sU16)_nstate;
   ////yac_host->printf("callMouseHandler: _mx=%i _my=%i absx=%i absy=%i cstate=%04x nstate=%04x grabbed=%i\n", _mx, _my, mouse.absx, mouse.absy, _cstate, _nstate, mouse.grabbed);
   if(mouse.grabbed)
   {
      mouse.dx = _mx - tks_HAL->view_sx / 2;
      mouse.dy = _my - tks_HAL->view_sy / 2;
      
      mouse.warp_sx = tks_HAL->view_sx / 2;
      mouse.warp_sy = tks_HAL->view_sy / 2;
      _mx = mouse.dx;
      _my = mouse.dy;
      mouse.absx += _mx;
      mouse.absy += _my;
      
      if(mouse.absx < 0) 
      {
         mouse.absx = 0;
      }
      else if(mouse.absx >= tks_HAL->view_sx)
      {
         mouse.absx = tks_HAL->view_sx-1;
      }
      if(mouse.absy < 0)
      {
         mouse.absy = 0;
      }
      else if(mouse.absy >= tks_HAL->view_sy)
      {
         mouse.absy = tks_HAL->view_sy-1;
      }
      _mx = mouse.absx;
      _my = mouse.absy;
   }
   else
   {
      mouse.dx = _mx-mouse.absx;
      mouse.dy = _my-mouse.absy;
      mouse.absx = _mx;
      mouse.absy = _my;
   }
   
   mouse.last_absx = mouse.dx;
   mouse.last_absy = mouse.dy;
   
   void *f=signal_funs[TKS_SIG_HALMOUSE];
   if(f)
   {
      YAC_Value args[4];
      args[0].initInt(_mx);
      args[1].initInt(_my);
      args[2].initInt(_cstate);
      args[3].initInt(_nstate);
      yac_host->yacEvalFunction(sc_context, f, 4, args);
   }
}

void _HAL::_reportJoystickEvents(sSI _b) {
   ::SDL_JoystickEventState(_b?SDL_ENABLE:SDL_DISABLE);
}

void _HAL::_delay(sSI _to) {
   ::SDL_Delay((sU32)_to);
}

sSI _HAL::_getTicks(void) {
   return ::SDL_GetTicks();
}

void _HAL::_setPriority(sSI _pri) {
   switch(_pri)
   {
   case TKS_PRIORITY_IDLE:
#ifdef YAC_WIN32
      SetPriorityClass(GetCurrentProcess(), IDLE_PRIORITY_CLASS);
      SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_IDLE);
#endif
      break;
   case TKS_PRIORITY_NORMAL:
#ifdef YAC_WIN32
      SetPriorityClass(GetCurrentProcess(), NORMAL_PRIORITY_CLASS);
      SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_NORMAL);
#endif
      break;
   case TKS_PRIORITY_HIGH:
#ifdef YAC_WIN32
      SetPriorityClass(GetCurrentProcess(), HIGH_PRIORITY_CLASS);
      SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_HIGHEST);
#endif
      break;
   case TKS_PRIORITY_REALTIME:
#ifdef YAC_WIN32
      SetPriorityClass(GetCurrentProcess(), REALTIME_PRIORITY_CLASS);
      SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_TIME_CRITICAL);
#endif
      break;
   }
}


void _HAL::_setRedrawFlag(sSI _fl) {
   redraw_flag=(sU8)_fl;
}

sSI _HAL::_getRedrawFlag(void) {
   return redraw_flag;
}

void _HAL::beginPaint(void) {
   if(!b_redrawing)
   {
      ::glViewport(0, 0, tks_HAL->view_sx, tks_HAL->view_sy);
      ::glMatrixMode(GL_TEXTURE);
      ::glLoadIdentity();
      ::glMatrixMode(GL_MODELVIEW);
      ::glLoadIdentity();
      b_redrawing=1;
      picking.first_tag=1;
      picking.rgb[0]=picking.rgb[1]=picking.rgb[2]=0;
      picking.z=1.0f;
      picking.retz=1.0f;
      //glDrawBuffer(GL_BACK);
   }
}

void _HAL::beginPaintDirty(void) {
   ::glViewport(0, 0, tks_HAL->view_sx, tks_HAL->view_sy);
   ::glMatrixMode(GL_TEXTURE);
   ::glLoadIdentity();
   ::glMatrixMode(GL_MODELVIEW);
   ::glLoadIdentity();
   b_redrawing=1;
   picking.first_tag=1;
   picking.rgb[0]=0;
   picking.rgb[1]=0;
   picking.rgb[2]=0;
   picking.z=1.0f;
   picking.retz=1.0f;
   //::glDrawBuffer(GL_FRONT);
}

void _HAL::endPaint(void) {
   if(b_redrawing)
   {
      if(b_graphics_active)
      {
         ::glFlush();
         ::SDL_GL_SwapBuffers();
      }
      b_redrawing=0;
   }
}

sBool _HAL::openWindow(sSI _sx, sSI _sy) {
   win_sx=(sU16)_sx;
   win_sy=(sU16)_sy;
   b_fullscreen=0;
   return openView(_sx, _sy, 0, 0, 0);
}

sBool _HAL::openScreen(sSI _sx, sSI _sy, sSI _z) {
   fs_sx=(sU16)_sx;
   fs_sy=(sU16)_sy;
   if(!win_sx)
   {
      win_sx=fs_sx;
      win_sy=fs_sy;
   }
   fs_z=(sU8)_z;
   b_fullscreen=1;
   return openView(_sx, _sy, _z, 1, 0);
}

sSI _HAL::openView(sU16 _sx, sU16 _sy, sU8 _z, 
                   sBool _fullscreen, sBool _resize) {
   if(!_resize)
   {
      // ---- close previously opened window ----
      closeView();
   }
   
   if(!::SDL_WasInit(SDL_INIT_VIDEO))
      ::SDL_InitSubSystem(SDL_INIT_VIDEO);
   
   view_sx=_sx;
   view_sy=_sy;
   
   sU32 flags;
   flags=SDL_OPENGL /*| SDL_RESIZABLE*/; 
   if(_fullscreen)
   {
      view_z=_z;
      flags|=SDL_FULLSCREEN;
   }
   else
   {
      const SDL_VideoInfo *info=::SDL_GetVideoInfo();
      if(!info)
      {
         yac_host->printf("\n[---] openView: SDL_GetVideoInfo failed.\n");
         return 0;
      }
      view_z=info->vfmt->BitsPerPixel;
      if(yac_host->yacGetDebugLevel()>7)
         yac_host->printf("\n[...] openView: desktop color depth=%ibit\n",
         info->vfmt->BitsPerPixel);
   }
   
   //if(!_resize)
   {
      ::SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 );   // require at least 5 bits per channel
      ::SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 5);
      ::SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5);
      ::SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16);
      ::SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, b_double_buffer);
   }
   SDL_Surface *surf=::SDL_SetVideoMode(view_sx, view_sy, view_z, flags|tks_flags);
   if( surf==NULL )
   {
      yac_host->printf("\n[---] SDL_SetVideoMode(%i, %i, %i, %08x) failed.\n", 
         view_sx, view_sy, view_z, flags);
      return 0;
   }
   // ---- workaround for SDL bug with very small windows (wrong surface geometry)
   int clipbox[4];
   //::glGetIntegerv(GL_SCISSOR_BOX, clipbox);  
   ::glGetIntegerv(GL_VIEWPORT, clipbox);  
   view_sx=surf->w;
   view_sy=surf->h;
   //if(clipbox[2]>view_sx) // old driver workaround
   {
      view_sx=clipbox[2];
   }
   //if(clipbox[3]>view_sy)
   {
      view_sy=clipbox[3];
   }
   ////yac_host->printf("[...] SDL_SetViewMode() OK: surface geo=(%i; %i) pitch=%i clipw=%i cliph=%i glh=%i glw=%i.\n", view_sx, view_sy, surf->pitch, surf->clip_rect.w, surf->clip_rect.h, clipbox[2], clipbox[3]);

   // ---- video mode is set
   if(_fullscreen)
   {
      view_flags|=_HAL::VP_FULLSCREEN;
   }
   view_flags|=_HAL::VP_OPENED;
   viewOpened();
   return 1;
}

void _HAL::closeView(void) {
   if(view_flags&VP_OPENED)
   {
      view_sx=0;
      view_sy=0;
      view_z=0;
      view_flags=0;
      b_graphics_active=0;
      if(::SDL_WasInit(SDL_INIT_VIDEO))
         ::SDL_QuitSubSystem(SDL_INIT_VIDEO);
#ifdef YAC_WIN32
      if(lpDD)
      {
         lpDD->Release();
         lpDD=0;
      }
#endif
   }
}

void _HAL::viewOpened(void) {
#ifdef YAC_WIN32
   if(lpDD)
   {
      lpDD->Release();
   }
   // ---- get monitor refresh rate ----
   ::DirectDrawCreate((LPGUID)0, &lpDD, NULL);
   lpDD->GetMonitorFrequency(&dw_monitorfreq);
   if(yac_host->yacGetDebugLevel()>7)
   {
      yac_host->printf("[...] Monitor frequency=%iHz\n", dw_monitorfreq);
   }
#endif
   
   resetMouse();
   b_graphics_active=1;
   
   glDisable(GL_BLEND);
   glDisable(GL_ALPHA_TEST);
   glDisable(GL_COLOR_LOGIC_OP);
   glDisable(GL_COLOR_MATERIAL);
   glDisable(GL_CULL_FACE);
   glDisable(GL_DEPTH_TEST);
   glDepthFunc(GL_LESS);
   glClearDepth(1.0f);
   glDisable(GL_DITHER);
   glDisable(GL_FOG);
   glDisable(GL_INDEX_LOGIC_OP);
   glDisable(GL_LIGHT0);
   glDisable(GL_LIGHT1);
   glDisable(GL_LIGHT2);
   glDisable(GL_LIGHT3);
   glDisable(GL_LIGHT4);
   glDisable(GL_LIGHT5);
   glDisable(GL_LIGHT6);
   glDisable(GL_LIGHT7);
   glDisable(GL_LIGHTING);
   glDisable(GL_LINE_SMOOTH);
   glDisable(GL_LINE_STIPPLE);
   glDisable(GL_MAP1_COLOR_4);
   glDisable(GL_MAP1_INDEX);
   glDisable(GL_MAP1_NORMAL);
   glDisable(GL_MAP1_TEXTURE_COORD_1);
   glDisable(GL_MAP1_TEXTURE_COORD_2);
   glDisable(GL_MAP1_TEXTURE_COORD_3);
   glDisable(GL_MAP1_TEXTURE_COORD_4);
   glDisable(GL_MAP1_VERTEX_3);
   glDisable(GL_MAP1_VERTEX_4);
   glDisable(GL_MAP2_COLOR_4);
   glDisable(GL_MAP2_INDEX);
   glDisable(GL_MAP2_NORMAL);
   glDisable(GL_MAP2_TEXTURE_COORD_1);
   glDisable(GL_MAP2_TEXTURE_COORD_2);
   glDisable(GL_MAP2_TEXTURE_COORD_3);
   glDisable(GL_MAP2_TEXTURE_COORD_4);
   glDisable(GL_MAP2_VERTEX_3);
   glDisable(GL_MAP2_VERTEX_4);
   glDisable(GL_NORMALIZE);
   glDisable(GL_POINT_SMOOTH);
   glDisable(GL_POLYGON_OFFSET_FILL);
   glDisable(GL_POLYGON_OFFSET_LINE);
   glDisable(GL_POLYGON_OFFSET_POINT);
   glDisable(GL_POLYGON_SMOOTH);
   glDisable(GL_POLYGON_STIPPLE);
   glDisable(GL_SCISSOR_TEST);
   glDisable(GL_STENCIL_TEST);
   glDisable(GL_TEXTURE_1D);
   glDisable(GL_TEXTURE_2D);
   glDisable(GL_TEXTURE_GEN_Q);
   glDisable(GL_TEXTURE_GEN_R);
   glDisable(GL_TEXTURE_GEN_S);
   glDisable(GL_TEXTURE_GEN_T);
   //glShadeModel(GL_FLAT);
   glShadeModel(GL_SMOOTH);
   glDrawBuffer(GL_BACK);
   glHint(GL_FOG_HINT, GL_NICEST);
   glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
 	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
   glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
   glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);

   
   // show GL driver info
   if(yac_host->yacGetDebugLevel()>7)
   {
      const GLubyte *gls=glGetString(GL_VENDOR);
      yac_host->printf("[...] OpenGL vendor    : \"%s\"\n", (char*)gls);
      gls=glGetString(GL_RENDERER);
      yac_host->printf("[...]        renderer  : \"%s\"\n", (char*)gls);
      gls=glGetString(GL_VERSION);
      yac_host->printf("[...]        version   : \"%s\"\n", (char*)gls);
      gls=glGetString(GL_EXTENSIONS);
      yac_host->printf("[...]        extensions: \"%s\"\n", (char*)gls);
   }

#ifdef TKSDL_SCRAP_CLIPBOARD
   init_scrap();
#endif

   initFPS();

}

sBool _HAL::toggleFullScreen(void) {
   sBool r=0;
   closeView();
   if(b_fullscreen)
      r=openWindow(win_sx, win_sy);
   else
   {
      YAC_Value cret;
      if(!(r=openScreen(fs_sx, fs_sy, fs_z)))
         r=openWindow(win_sx, win_sy);
   }
   callReopen();
   return r;
}

void _HAL::useWindow(sU16 _sx, sU16 _sy) {
   win_sx=(sU16)_sx;
   win_sy=(sU16)_sy;
   view_sx=_sx;
   view_sy=_sy;
   
}

void _HAL::waitVBlank(void) {
#ifdef YAC_WIN32
   if(lpDD)
   {
      lpDD->WaitForVerticalBlank(DDWAITVB_BLOCKBEGIN, NULL);
   }
#endif
}

sUI _HAL::getMonitorFrequency(void) {
#ifdef YAC_WIN32
   if(lpDD)
   {
      lpDD->GetMonitorFrequency(&dw_monitorfreq);
      return dw_monitorfreq;
   }
   yac_host->printf("[...] Viewport::getFrequency: no ddraw pointer set, no window/screen opened?\n");
#endif
   return 60; // xxx default frequency
}

sF32 _HAL::getPickedZ(void) {
   if(picking.first_tag)
   {
      picking.first_tag=0;
      
      if(picking.z==1.0f)
      {
         picking.retz=1.0f;
         picking.rgb[0]=0;
         picking.rgb[1]=0;
         picking.rgb[2]=0;
      }
      else
      {
         //yac_host->printf("picking.z=%f, picking.retz=%f\n", picking.z, picking.retz);
         picking.retz=(picking.retz+1.0f)/2.0f;
         if(picking.retz>1.0f)
            picking.retz=1.0f;
         else
            if(picking.retz<0.0f)
               picking.retz=0.0f;
      }
   }
   return picking.retz;
}

sUI _HAL::getPickedID(void) {
   return (picking.rgb[1]<<8)|(picking.rgb[2]);
}

void _HAL::_setEventPolling(sSI _b) {
   b_event_polling=(_b!=0);
}

sSI _HAL::_getEventPolling(void) {
   return b_event_polling;
}

void _HAL::showCursor(sBool _st) {
   if(mouse.show!=_st)
   {
      mouse.show=_st; 
      ::SDL_ShowCursor(_st?SDL_ENABLE:SDL_DISABLE);
   }
}

sSI _HAL::_enableUNICODE(sSI _bEnabled) {
   if(_bEnabled>=0)
      b_unicode=_bEnabled;
   return ::SDL_EnableUNICODE(_bEnabled);
}

void _HAL::updateUnicode(void) {
   ///yac_host->printf("updateUniCode(%i)=%i\n", b_unicode, ::SDL_EnableUNICODE(b_unicode));
   ::SDL_EnableUNICODE(b_unicode);
}

sSI _HAL::_isApplicationActive(void) {
   return b_graphics_active;
}

sSI _HAL::_getSysWMInfo(void) {
   SDL_VERSION(&syswminfo.version);
   if(::SDL_GetWMInfo(&syswminfo))
   {
      return (sSI)(&syswminfo);
   }
   else
   {
      return 0;
   }
}

/*sSI _HAL::_getSysWindowHandle(void) {
   if(::SDL_GetWMInfo(&syswminfo))
   {
#ifdef YAC_WIN32
      return (sSI)(syswminfo.window);
#else
      return (sSI)(syswminfo.x11.window);
#endif
   }
   else
   {
      return 0;
   }
}
*/
